﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using System.Windows.Forms;
using Windows;

namespace Widow
{
    public partial class RuleEditForm : Form
    {
        #region Public Enums, Properties and Fields

        public Windows.Windows Windows { get; }

        public MainForm Form { get; set; }

        #endregion

        #region Constructors, Destructors and Finalizers

        public RuleEditForm(MainForm mainForm, Windows.Windows windows) : this(windows)
        {
            Form = mainForm;

            Form.WindowCreated += Form_WindowCreated;
            Form.WindowDestroyed += Form_WindowDestroyed;
        }

        public RuleEditForm(Windows.Windows windows) : this()
        {
            Windows = windows;

            foreach (var window in windows.Window)
            {
                windowRulesListBox.Items.Add(window);
            }
        }

        private RuleEditForm()
        {
            InitializeComponent();

            desktopWindowsListBox.DisplayMember = nameof(Window.Title);
            desktopWindowsListBox.SetDoubleBuffered();
            windowRulesListBox.DisplayMember = nameof(Window.Title);
            windowRulesListBox.SetDoubleBuffered();

            Task.Run(RefreshWindows);
        }

        /// <summary>
        ///     Clean up any resources being used.
        /// </summary>
        /// <param name="disposing">true if managed resources should be disposed; otherwise, false.</param>
        protected override void Dispose(bool disposing)
        {
            if (disposing && components != null)
            {
                Form.WindowCreated -= Form_WindowCreated;
                Form.WindowDestroyed -= Form_WindowDestroyed;

                components.Dispose();
            }

            base.Dispose(disposing);
        }

        #endregion

        #region Event Handlers

        private void Form_WindowDestroyed(object sender, WindowDestroyedEventArgs e)
        {
            if (string.IsNullOrEmpty(e.Title))
            {
                return;
            }

            var indices = new List<int>();
            for (var i = 0; i < desktopWindowsListBox.Items.Count; ++i)
            {
                if (((Window) desktopWindowsListBox.Items[i]).Title == e.Title)
                {
                    indices.Add(i);
                }
            }

            foreach (var index in indices)
            {
                desktopWindowsListBox.Items.RemoveAt(index);
            }
        }

        private void Form_WindowCreated(object sender, WindowCreatedEventArgs e)
        {
            if (string.IsNullOrEmpty(e.Title))
            {
                return;
            }

            var found = false;
            foreach (var window in desktopWindowsListBox.Items)
            {
                if (((Window) window).Title == e.Title)
                {
                    found = true;
                    break;
                }
            }

            if (found)
            {
                return;
            }

            if (!Natives.GetWindowRect(e.Handle, out var rect))
            {
                return;
            }

            var process = Helpers.GetProcessName(e.Handle);
            if (string.IsNullOrEmpty(process))
            {
                return;
            }

            var newWindow = new Window(process, e.Title, rect.Top, rect.Left, rect.Right - rect.Left,
                rect.Bottom - rect.Top);

            desktopWindowsListBox.Items.Add(newWindow);
        }

        private void WindowRulesListBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            var listBox = (ListBox) sender;

            if (listBox.SelectedIndex == -1)
            {
                WindowTitle.Text = string.Empty;
                ignoreLeftCheckBox.Checked = false;
                WindowLeft.Text = string.Empty;
                ignoreTopCheckBox.Checked = false;
                WindowTop.Text = string.Empty;
                ignoreWidthCheckBox.Checked = false;
                WindowWidth.Text = string.Empty;
                ignoreHeightCheckBox.Checked = false;
                WindowHeight.Text = string.Empty;
                return;
            }

            var window = (Window) listBox.SelectedItem;

            if (window == null)
            {
                return;
            }

            WindowTitle.Text = window.Title;
            ignoreLeftCheckBox.Checked = window.IgnoreLeft;
            WindowLeft.Text = window.Left.ToString();
            ignoreTopCheckBox.Checked = window.IgnoreTop;
            WindowTop.Text = window.Top.ToString();
            ignoreWidthCheckBox.Checked = window.IgnoreWidth;
            WindowWidth.Text = window.Width.ToString();
            ignoreHeightCheckBox.Checked = window.IgnoreHeight;
            WindowHeight.Text = window.Height.ToString();
        }

        private void RefreshButton_Click(object sender, EventArgs e)
        {
            Task.Run(RefreshWindows);
        }

        private void RemoveButton_Click(object sender, EventArgs e)
        {
            var window = (Window) windowRulesListBox.SelectedItem;
            if (window == null)
            {
                return;
            }

            var index = -1;
            for (var i = 0; i < windowRulesListBox.Items.Count; ++i)
            {
                if (((Window) windowRulesListBox.Items[i]).Title != window.Title)
                {
                    continue;
                }

                index = i;
                break;
            }

            if (index == -1)
            {
                return;
            }

            windowRulesListBox.Items.RemoveAt(index);
            Windows.Window.Remove(window);
        }

        private void AddButton_Click(object sender, EventArgs e)
        {
            var window = (Window) desktopWindowsListBox.SelectedItem;
            if (window == null)
            {
                return;
            }

            var found = false;
            foreach (var windowRule in windowRulesListBox.Items)
            {
                if (((Window) windowRule).Title != window.Title)
                {
                    continue;
                }

                found = true;
                break;
            }

            if (found)
            {
                return;
            }

            windowRulesListBox.Items.Add(window);
            Windows.Window.Add(window);
        }

        private void OnWindowSettings_TextChanged(object sender, EventArgs e)
        {
            var selectedWindow = (Window) windowRulesListBox.SelectedItem;

            if (selectedWindow == null)
            {
                WindowTitle.Text = string.Empty;
                WindowLeft.Text = string.Empty;
                WindowTop.Text = string.Empty;
                WindowWidth.Text = string.Empty;
                WindowHeight.Text = string.Empty;

                return;
            }

            var textBox = (TextBox) sender;
            switch (textBox.Name)
            {
                case nameof(WindowTitle):
                    selectedWindow.Title = textBox.Text;
                    break;
                case nameof(WindowLeft):
                    if (int.TryParse(textBox.Text, out var left))
                    {
                        selectedWindow.Left = left;
                    }

                    break;
                case nameof(WindowTop):
                    if (int.TryParse(textBox.Text, out var top))
                    {
                        selectedWindow.Top = top;
                    }

                    break;
                case nameof(WindowWidth):
                    if (int.TryParse(textBox.Text, out var width))
                    {
                        selectedWindow.Width = width;
                    }

                    break;
                case nameof(WindowHeight):
                    if (int.TryParse(textBox.Text, out var height))
                    {
                        selectedWindow.Height = height;
                    }

                    break;
            }
        }

        private void OnIgnoreWindowSettings_CheckedChanged(object sender, EventArgs e)
        {
            var selectedWindow = (Window) windowRulesListBox.SelectedItem;

            if (selectedWindow == null)
            {
                ignoreLeftCheckBox.Checked = false;
                ignoreTopCheckBox.Checked = false;
                ignoreWidthCheckBox.Checked = false;
                ignoreHeightCheckBox.Checked = false;

                return;
            }

            var checkBox = (CheckBox) sender;
            switch (checkBox.Name)
            {
                case nameof(ignoreHeightCheckBox):
                    selectedWindow.IgnoreHeight = checkBox.Checked;
                    break;
                case nameof(ignoreWidthCheckBox):
                    selectedWindow.IgnoreWidth = checkBox.Checked;
                    break;
                case nameof(ignoreTopCheckBox):
                    selectedWindow.IgnoreTop = checkBox.Checked;
                    break;
                case nameof(ignoreLeftCheckBox):
                    selectedWindow.IgnoreLeft = checkBox.Checked;
                    break;
            }
        }

        #endregion

        #region Private Methods

        private void RefreshWindows()
        {
            foreach (var handle in Helpers.FindWindows((wnd, param) => true))
            {
                var title = Helpers.GetWindowTitle(handle);
                if (string.IsNullOrEmpty(title))
                {
                    continue;
                }

                var process = Helpers.GetProcessName(handle);
                if (string.IsNullOrEmpty(process))
                {
                    continue;
                }

                if (!Natives.GetWindowRect(handle, out var rect))
                {
                    continue;
                }

                var window = new Window(process, title, rect.Top, rect.Left, rect.Right - rect.Left,
                    rect.Bottom - rect.Top);

                this.Execute(() =>
                {
                    var found = false;
                    foreach (var item in desktopWindowsListBox.Items)
                    {
                        if (((Window) item).Title != window.Title)
                        {
                            continue;
                        }

                        found = true;
                        break;
                    }

                    if (!found)
                    {
                        desktopWindowsListBox.Items.Add(window);
                    }
                });
            }
        }

        #endregion
    }
}